@@ -1,23 +1,24 @@
 """Automates poetry operations in the repo."""
 
-import re
 from collections import defaultdict
+from importlib.metadata import version as package_version
 from pathlib import Path
-from typing import Generator, Iterable, List, Set, Union
+from typing import Generator, Optional, Set, Tuple, Union
 
-import click
+from cleo.io.io import IO
 from coveo_functools.finalizer import finalizer
-from coveo_styles.styles import ExitWithFailure, echo, install_pretty_exception_hook
+from coveo_styles.styles import ExitWithFailure, echo
 from coveo_systools.filesystem import find_repo_root
 
 from coveo_stew.ci.runner_status import RunnerStatus
-from coveo_stew.discovery import discover_pyprojects, find_pyproject
+from coveo_stew.discovery import Predicate, discover_pyprojects, find_pyproject
 from coveo_stew.exceptions import (
     CheckFailed,
     PythonProjectNotFound,
     RequirementsOutdated,
 )
 from coveo_stew.offline_publish import offline_publish
+from coveo_stew.presets.presets_list import STEW_PRESETS_LIST
 from coveo_stew.pydev import is_pydev_project, pull_and_write_dev_requirements
 from coveo_stew.stew import (
     EnvironmentCreationBehavior,
@@ -25,8 +26,6 @@ from coveo_stew.stew import (
     PythonProject,
 )
 
-_COMMANDS_THAT_SKIP_INTRO_EMOJIS = ["locate"]
-
 
 def _echo_updated(updated: Set[Path]) -> None:
     """Used to print updated paths to the user."""
@@ -43,78 +42,104 @@ def _echo_updated(updated: Set[Path]) -> None:
 
 
 def _pull_dev_requirements(
-    dry_run: bool = False, verbose: bool = False
+    io: IO, dry_run: bool = False, verbose: bool = False, disable_cache: bool = False
 ) -> Generator[Path, None, None]:
     """Writes the dev-dependencies of pydev projects' local dependencies into pydev's pyproject.toml file."""
     dry_run_text = "(dry run) " if dry_run else ""
-    for pydev_project in discover_pyprojects(predicate=is_pydev_project, verbose=verbose):
+    for pydev_project in _discover_pyprojects(
+        io, predicate=is_pydev_project, verbose=verbose, disable_cache=disable_cache
+    ):
         echo.step(f"Analyzing dev requirements for {pydev_project}")
-        if pull_and_write_dev_requirements(pydev_project, dry_run=dry_run):
+        if pull_and_write_dev_requirements(io, pydev_project, dry_run=dry_run):
             echo.outcome(
-                f"{dry_run_text}Updated {pydev_project.package.name} with new dev requirements."
+                f"{dry_run_text}Updated {pydev_project.poetry.package.pretty_name} with new dev requirements."
             )
             if not dry_run:
                 echo.outcome("Lock file and virtual environment updated !!thumbs_up!!\n")
-            yield pydev_project.toml_path
+            yield pydev_project.poetry.pyproject_path
         else:
-            echo.success(f"{pydev_project.package.name}'s dev requirements were up to date.")
+            echo.success(
+                f"{pydev_project.poetry.package.pretty_name}'s dev requirements were up to date."
+            )
 
 
-@click.group()
-@click.pass_context
-def stew(ctx: click.Context) -> None:
-    """The 'stew' cli entry point."""
-    install_pretty_exception_hook()
-    if ctx.invoked_subcommand not in _COMMANDS_THAT_SKIP_INTRO_EMOJIS:
-        echo.step("!!sparkles!! !!snake!! !!sparkles!!")
+def version(io: IO) -> None:
+    """Prints the version of the coveo-stew package."""
+    io.write_line(f"coveo-stew {package_version('coveo-stew')}")
 
 
-@stew.command()
-@click.option("--verbose", is_flag=True, default=False)
-def check_outdated(verbose: bool = False) -> None:
+def check_outdated(
+    io: IO,
+    project_name: Optional[str] = None,
+    exact_match: bool = False,
+    verbose: bool = False,
+    disable_cache: bool = False,
+) -> None:
     """Return error code 1 if toml/lock are not in sync."""
-    echo.step("Analyzing all pyproject.toml files and artifacts:")
+    echo.step("Analyzing pyproject.toml files and artifacts:")
     outdated: Set[Path] = set()
     try:
-        for project in discover_pyprojects(verbose=verbose):
+        for project in _discover_pyprojects(
+            io,
+            query=project_name,
+            exact_match=exact_match,
+            verbose=verbose,
+            disable_cache=disable_cache,
+        ):
             echo.noise(project, item=True)
-            if not project.lock_path.exists() or project.lock_is_outdated():
-                outdated.add(project.lock_path)
+            if not (project.poetry.locker.is_locked() and project.poetry.locker.is_fresh()):
+                outdated.add(project.poetry.locker.lock)
     except PythonProjectNotFound as exception:
         raise ExitWithFailure from exception
 
     try:
-        outdated.update(_pull_dev_requirements(dry_run=True, verbose=verbose))
+        outdated.update(
+            _pull_dev_requirements(io, dry_run=True, verbose=verbose, disable_cache=disable_cache)
+        )
     except PythonProjectNotFound:
         pass  # no pydev projects found.
 
     if outdated:
         raise ExitWithFailure(
             failures=outdated,
-            suggestions='Run "poetry run pyproject fix-outdated" to update all outdated files.',
+            suggestions='Run "stew fix-outdated" to update all outdated files.',
         ) from RequirementsOutdated(f"Found {len(outdated)} outdated file(s).")
 
-    echo.success("Check complete! All files are up-to-date.")
+    echo.success("Check complete! All scanned files are up-to-date.")
 
 
-@stew.command()
-@click.option("--verbose", is_flag=True, default=False)
-def fix_outdated(verbose: bool = False) -> None:
+def fix_outdated(
+    io: IO,
+    project_name: Optional[str] = None,
+    exact_match: bool = False,
+    verbose: bool = False,
+    disable_cache: bool = False,
+) -> None:
     """Scans the whole repo and updates outdated pyproject-related files.
 
     Updates:
         - Lock files, only if their pyproject.toml was updated.
     """
-    echo.step("Synchronizing all outdated lock files:")
+    echo.step("Synchronizing outdated lock files:")
     updated: Set[Path] = set()
     with finalizer(_echo_updated, updated):
         try:
-            for project in discover_pyprojects(verbose=verbose):
+            for project in _discover_pyprojects(
+                io,
+                query=project_name,
+                exact_match=exact_match,
+                verbose=verbose,
+                disable_cache=disable_cache,
+            ):
                 echo.noise(project, item=True)
                 if project.lock_if_needed():
-                    updated.add(project.lock_path)
+                    updated.add(project.poetry.locker.lock)
             try:
-                updated.update(_pull_dev_requirements(dry_run=False, verbose=verbose))
+                updated.update(
+                    _pull_dev_requirements(
+                        io, dry_run=False, verbose=verbose, disable_cache=disable_cache
+                    )
+                )
             except PythonProjectNotFound:
                 pass  # no pydev projects found
         except PythonProjectNotFound as exception:
@@ -123,33 +148,42 @@ def fix_outdated(verbose: bool = False) -> None:
     echo.success(f'Update complete! {len(updated) or "No"} file(s) were modified.\n')
 
 
-@stew.command()
-@click.option("--verbose", is_flag=True, default=False)
-def bump(verbose: bool = False) -> None:
-    """Bumps locked versions for all pyprojects."""
+def bump(
+    io: IO,
+    project_name: Optional[str] = None,
+    exact_match: bool = False,
+    verbose: bool = False,
+    disable_cache: bool = False,
+) -> None:
+    """Bumps locked versions."""
     updated: Set[Path] = set()
     with finalizer(_echo_updated, updated):
         try:
-            for project in discover_pyprojects(verbose=verbose):
-                echo.step(f"Bumping {project.lock_path}")
+            for project in _discover_pyprojects(
+                io,
+                query=project_name,
+                exact_match=exact_match,
+                verbose=verbose,
+                disable_cache=disable_cache,
+            ):
+                echo.noise(project, item=True)
+                echo.step(f"Bumping {project.poetry.locker.lock}")
                 if project.bump():
-                    updated.add(project.toml_path)
+                    updated.add(project.poetry.pyproject_path)
         except PythonProjectNotFound as exception:
             raise ExitWithFailure from exception
 
     echo.success(f'Bump complete! {len(updated) or "No"} file(s) were modified.')
 
 
-@stew.command()
-@click.argument("project_name")
-@click.option("--directory", default=None)
-@click.option("--python", default=None)
-@click.option("--verbose", is_flag=True, default=False)
 def build(
-    project_name: str,
+    io: IO,
+    project_name: Optional[str] = None,
+    exact_match: bool = True,
     directory: Union[str, Path] = None,
     python: Union[str, Path] = None,
     verbose: bool = False,
+    disable_cache: bool = False,
 ) -> None:
     """
     Store all dependencies of a python project into a local directory, according to its poetry.lock,
@@ -160,11 +194,31 @@ def build(
         IF unspecified and no repo: "pyproject_folder/.wheels/*.whl"
         IF specified:               "directory/*.whl"
     """
+    if not project_name:
+        exact_match = False  # if you write `stew build` we build all.
+
     try:
-        project = find_pyproject(project_name, verbose=verbose)
+        for project in _discover_pyprojects(
+            io,
+            query=project_name,
+            exact_match=exact_match,
+            verbose=verbose,
+            disable_cache=disable_cache,
+        ):
+            echo.noise(project, item=True)
+            _build(io, project, directory, python)
     except PythonProjectNotFound as exception:
         raise ExitWithFailure from exception
 
+    echo.success()
+
+
+def _build(
+    io: IO,
+    project: PythonProject,
+    directory: Optional[Union[str, Path]],
+    python: Optional[Union[str, Path]],
+) -> None:
     python_environments = (
         [PythonEnvironment(python)]
         if python
@@ -176,20 +230,22 @@ def build(
     if not directory:
         directory = (project.repo_root or project.project_path) / ".wheels"
     assert directory
-    directory = Path(directory)
+    directory = Path(directory).absolute()
 
     echo.step(f"Building python project {project} in {directory}")
     for environment in python_environments:
+        project.install(environment=environment)
         echo.outcome(f"virtual environment: {environment}", pad_before=True)
-        offline_publish(project, directory, environment)
-
-    echo.success()
+        offline_publish(io, project, directory, environment)
 
 
-@stew.command()
-@click.argument("project_name", default=None, required=False)
-@click.option("--verbose", is_flag=True, default=False)
-def fresh_eggs(project_name: str = None, verbose: bool = False) -> None:
+def fresh_eggs(
+    io: IO,
+    project_name: str = None,
+    exact_match: bool = False,
+    verbose: bool = False,
+    disable_cache: bool = False,
+) -> None:
     """
     Removes the egg-info from project folders.
 
@@ -204,8 +260,16 @@ def fresh_eggs(project_name: str = None, verbose: bool = False) -> None:
     """
     echo.step("Removing *.egg-info folders.")
     deleted = False
+
     try:
-        for project in discover_pyprojects(query=project_name, verbose=verbose):
+        for project in _discover_pyprojects(
+            io,
+            query=project_name,
+            verbose=verbose,
+            exact_match=exact_match,
+            disable_cache=disable_cache,
+        ):
+            echo.noise(project, item=True)
             if project.remove_egg_info():
                 echo.outcome("Deleted: ", project.egg_path, item=True)
                 deleted = True
@@ -218,54 +282,31 @@ def fresh_eggs(project_name: str = None, verbose: bool = False) -> None:
     echo.success()
 
 
-@stew.command()
-@click.option("--dry-run/--no-dry-run", default=False)
-@click.option("--verbose", is_flag=True, default=False)
-def pull_dev_requirements(dry_run: bool = False, verbose: bool = False) -> None:
+def pull_dev_requirements(
+    io: IO, dry_run: bool = False, verbose: bool = False, disable_cache: bool = False
+) -> None:
     """Writes the dev-dependencies of pydev projects' local dependencies into pydev's pyproject.toml file."""
     try:
-        list(_pull_dev_requirements(dry_run=dry_run, verbose=verbose))
+        list(
+            _pull_dev_requirements(
+                io, dry_run=dry_run, verbose=verbose, disable_cache=disable_cache
+            )
+        )
     except PythonProjectNotFound as exception:
         raise ExitWithFailure from exception
 
 
-def _beautify_mypy_output(
-    project: PythonProject, output: Iterable[str], *, full_paths: bool = False
-) -> None:
-    """Main use: guide IDEs by showing full paths to the files vs the current working directory.
-    Bonus: highlight errors in red and display a slightly shortened version of the error output."""
-    pattern = re.compile(
-        rf"^(?P<path>{project.package.safe_name}.+):(?P<line>\d+):(?P<column>\d+(?::)| )"
-        rf"(?:\s?error:\s?)(?P<detail>.+)$"
-    )
-    for line in output:
-        match = pattern.fullmatch(line)
-        if match:
-            adjusted_path = project.project_path / Path(match["path"])
-            adjusted_path = (
-                adjusted_path.resolve()
-                if full_paths
-                else adjusted_path.relative_to(Path(".").resolve())
-            )
-            echo.error_details(
-                f'{adjusted_path}:{match["line"]}:{match["column"]} {match["detail"]}'
-            )
-        else:
-            echo.noise(line)
-
-
-@stew.command()
-@click.argument("project_name")
-@click.option("--verbose", is_flag=True, default=False)
-def locate(project_name: str, verbose: bool = False) -> None:
+def locate(io: IO, project_name: str, verbose: bool = False, disable_cache: bool = False) -> None:
     """Locate a python project (in the whole git repo) and print the directory containing the pyproject.toml file."""
     try:
-        echo.passthrough(find_pyproject(project_name, verbose=verbose).project_path)
+        echo.passthrough(find_pyproject(io, project_name, verbose=verbose).project_path)
     except PythonProjectNotFound as exception:
         # check for partial matches to guide the user
         partial_matches = (
-            project.package.name
-            for project in discover_pyprojects(query=project_name, verbose=verbose)
+            project.poetry.package.pretty_name
+            for project in _discover_pyprojects(
+                io, query=project_name, verbose=verbose, disable_cache=disable_cache
+            )
         )
         try:
             raise ExitWithFailure(
@@ -279,16 +320,23 @@ def locate(project_name: str, verbose: bool = False) -> None:
             raise ExitWithFailure from exception
 
 
-@stew.command()
-@click.argument("project_name", default=None, required=False)
-@click.option("--exact-match/--no-exact-match", default=False)
-@click.option("--verbose", is_flag=True, default=False)
-def refresh(project_name: str = None, exact_match: bool = False, verbose: bool = False) -> None:
+def refresh(
+    io: IO,
+    project_name: str = None,
+    exact_match: bool = False,
+    verbose: bool = False,
+    disable_cache: bool = False,
+) -> None:
+    """Refresh python project environments."""
     echo.step("Refreshing python project environments...")
     pydev_projects = []
     try:
-        for project in discover_pyprojects(
-            query=project_name, exact_match=exact_match, verbose=verbose
+        for project in _discover_pyprojects(
+            io,
+            query=project_name,
+            exact_match=exact_match,
+            verbose=verbose,
+            disable_cache=disable_cache,
         ):
             if project.options.pydev:
                 pydev_projects.append(project)
@@ -308,38 +356,40 @@ def refresh(project_name: str = None, exact_match: bool = False, verbose: bool =
     echo.success()
 
 
-@stew.command()
-@click.argument("project_name", default=None, required=False)
-@click.option("--exact-match/--no-exact-match", default=False)
-@click.option("--fix/--no-fix", default=False)
-@click.option("--check", multiple=True, default=None)
-@click.option("--skip", multiple=True, default=None)
-@click.option("--verbose", is_flag=True, default=False)
-@click.option(
-    "--quick",
-    is_flag=True,
-    default=False,
-    help="Do not call 'poetry install --remove-untracked' before testing.",
-)
-@click.option("--parallel/--sequential", default=True)
-@click.option("--github-step-report", is_flag=True, default=False, envvar="GITHUB_ACTIONS")
 def ci(
+    io: IO,
     project_name: str = None,
     exact_match: bool = False,
     fix: bool = False,
-    check: List[str] = None,
-    skip: List[str] = None,
+    check: Tuple[str, ...] = (),
+    skip: Tuple[str, ...] = (),
     verbose: bool = False,
     quick: bool = False,
     parallel: bool = True,
     github_step_report: bool = False,
+    extra: Tuple[str, ...] = (),
+    no_extras: bool = False,
+    all_extras: bool = False,
+    disable_cache: bool = False,
 ) -> None:
+    """Run continuous integration steps on Python projects."""
     failures = defaultdict(list)
     try:
-        for project in discover_pyprojects(
-            query=project_name, exact_match=exact_match, verbose=verbose
+        for project in _discover_pyprojects(
+            io,
+            query=project_name,
+            exact_match=exact_match,
+            verbose=verbose,
+            disable_cache=disable_cache,
         ):
-            echo.step(project.package.name, pad_after=False)
+            echo.step(project.poetry.package.pretty_name, pad_after=False)
+
+            if quick:
+                check += tuple(project.options.quick.get("check", ()))
+                skip += tuple(project.options.quick.get("skip", ()))
+
+            project.overrides_from_cli(extras=extra, no_extras=no_extras, all_extras=all_extras)
+
             if (
                 overall_result := project.launch_continuous_integration(
                     auto_fix=fix,
@@ -362,3 +412,44 @@ def ci(
         raise ExitWithFailure(failures=projects, exit_code=exit_code) from CheckFailed(
             f"{len(failures)} project(s) failed ci steps."
         )
+
+
+def _discover_pyprojects(
+    io: IO,
+    query: Optional[str] = None,
+    predicate: Optional[Predicate] = None,
+    exact_match: bool = False,
+    verbose: bool = False,
+    disable_cache: bool = False,
+) -> Generator[PythonProject, None, None]:
+    if query and query.startswith("."):
+        if exact_match:
+            echo.warning(
+                f"--exact-match only works with project names, but we are targeting a path: {query}"
+            )
+        yield from discover_pyprojects(
+            io,
+            path=Path(query),
+            verbose=verbose,
+            find_nested=False,
+            predicate=predicate,
+            disable_cache=disable_cache,
+        )
+    else:
+        yield from discover_pyprojects(
+            io,
+            query=query,
+            exact_match=exact_match,
+            verbose=verbose,
+            predicate=predicate,
+            disable_cache=disable_cache,
+        )
+
+
+def presets_list(io: IO) -> None:
+    """List the builtin stew presets."""
+    io.write_line("\n<fg=green>Presets:</>")
+    for preset_fn, description in STEW_PRESETS_LIST:
+        preset_name = preset_fn.__name__.replace("_", "-")
+        io.write_line(f"  <fg=yellow>{preset_name:<20}</>: {description}")
+    io.write_line("")
