@@ -15,6 +15,7 @@ from typing import (
     Any,
     Final,
     Generator,
+    Iterable,
     Iterator,
     List,
     Optional,
@@ -23,18 +24,20 @@ from typing import (
     Union,
 )
 
+from cleo.io.io import IO
+from coveo_functools import flex
 from coveo_functools.casing import flexfactory
-from coveo_itertools.lookups import dict_lookup
 from coveo_systools.filesystem import CannotFindRepoRoot, find_repo_root
 from coveo_systools.subprocess import check_run
+from poetry.factory import Factory
+from poetry.poetry import Poetry
 
 from coveo_stew.ci.runner_status import RunnerStatus
+from coveo_stew.config import load_config_from_presets
 from coveo_stew.environment import PythonEnvironment, PythonTool, find_python_tool
-from coveo_stew.exceptions import NotAPoetryProject, StewException
-from coveo_stew.metadata.poetry_api import PoetryAPI
-from coveo_stew.metadata.python_api import PythonFile
+from coveo_stew.exceptions import NotAPoetryProject, StewException, UsageError
 from coveo_stew.metadata.stew_api import StewPackage
-from coveo_stew.poetry_backward_compatibility import get_verb
+from coveo_stew.poetry_backward_compatibility import get_install_sync_command
 from coveo_stew.utils import load_toml_from_path
 
 ENVIRONMENT_PATH_PATTERN: Final[Pattern] = re.compile(
@@ -42,10 +45,6 @@ ENVIRONMENT_PATH_PATTERN: Final[Pattern] = re.compile(
 )
 
 
-# temporary until 1.2.0 has been cleared as fully compatible
-_CHECKED_POETRY_VERSION: bool = False
-
-
 class EnvironmentCreationBehavior(Enum):
     Full = auto()
     NoDev = auto()
@@ -55,43 +54,57 @@ class EnvironmentCreationBehavior(Enum):
 class PythonProject:
     """Access the information within a pyproject.toml file and operate on them."""
 
-    def __init__(self, project_path: Path, *, verbose: bool = False) -> None:
+    def __init__(
+        self,
+        io: IO,
+        poetry: Union[Poetry, Path],
+        *,
+        verbose: bool = False,
+        disable_cache: bool = False,
+    ) -> None:
+        self.io = io
         self.verbose = verbose
-        self.project_path: Path = (
-            project_path if project_path.is_dir() else project_path.parent
-        ).absolute()
-        self.toml_path: Path = self.project_path / PythonFile.PyProjectToml
-        self.lock_path: Path = self.project_path / PythonFile.PoetryLock
 
-        toml_content = load_toml_from_path(self.toml_path)
+        if isinstance(poetry, Path):
+            try:
+                self.poetry = Factory().create_poetry(
+                    cwd=poetry.parent if poetry.is_file() else poetry,
+                    io=io,
+                    disable_plugins=False,
+                    disable_cache=disable_cache,
+                )
+            except RuntimeError as ex:
+                raise NotAPoetryProject(poetry) from ex
+        else:
+            self.poetry = poetry
 
-        try:
-            self.package: PoetryAPI = flexfactory(
-                PoetryAPI, **dict_lookup(toml_content, "tool", "poetry")
-            )
-        except KeyError as exception:
-            raise NotAPoetryProject from exception
+        self.project_path = self.poetry.pyproject_path.parent
 
-        self.egg_path: Path = self.project_path / f"{self.package.safe_name}.egg-info"
+        self.dependencies = set(self.poetry.package.requires)
+        self.all_dependencies = set(self.poetry.package.all_requires)
+        self.dev_dependencies = self.all_dependencies - self.dependencies
 
-        self.options: StewPackage = flexfactory(
-            StewPackage,
-            **dict_lookup(toml_content, "tool", "stew", default={}),
-            _pyproject=self,
-        )
+        toml_content = load_toml_from_path(self.poetry.pyproject_path)
+
+        self.egg_path: Path = self.project_path / f"{self.poetry.package.name}.egg-info"
+
+        stew_config, ci_config = load_config_from_presets(io, toml_content)
+        self.options = flex.deserialize(stew_config, hint=StewPackage, errors="raise")
 
         from coveo_stew.ci.config import ContinuousIntegrationConfig  # circular import
 
         if self.options.pydev:
             # ensure no steps are repeated. pydev projects only receive basic poetry/lock checks
             self.ci: ContinuousIntegrationConfig = ContinuousIntegrationConfig(
-                check_outdated=True, poetry_check=True, mypy=False, _pyproject=self
+                check_outdated=True,
+                poetry_check=True,
+                mypy=False,
+                _pyproject=self,
+                io=self.io,
             )
         else:
             self.ci = flexfactory(
-                ContinuousIntegrationConfig,
-                **dict_lookup(toml_content, "tool", "stew", "ci", default={}),
-                _pyproject=self,
+                ContinuousIntegrationConfig, **ci_config, io=self.io, _pyproject=self
             )
 
         try:
@@ -116,19 +129,37 @@ class PythonProject:
 
         return cache
 
+    def overrides_from_cli(
+        self, extras: Tuple[str, ...] = (), no_extras: bool = False, all_extras: bool = False
+    ) -> None:
+        """Overrides the project's options with the CLI arguments."""
+        if no_extras and (extras or all_extras):
+            raise UsageError("Cannot use --no-extras with --extras or --all-extras.")
+        if all_extras and (extras or no_extras):
+            raise UsageError("Cannot use --all-extras with --extras or --no-extras.")
+
+        if extras:
+            self.options.extras = list(extras)
+            self.options.all_extras = False
+        if no_extras:
+            self.options.extras = []
+            self.options.all_extras = False
+        if all_extras:
+            self.options.extras = []
+            self.options.all_extras = True
+
     def relative_path(self, path: Path) -> Path:
         """returns the relative path of a path vs the project folder."""
         return path.relative_to(self.project_path)
 
     def lock_is_outdated(self) -> bool:
         """True if the toml file has pending changes that were not applied to poetry.lock"""
-        if not self.lock_path.exists():
+        if not self.poetry.locker.is_locked():
             return False
 
         # yolo: use the dry run output to determine if the lock is too old
         dry_run_output = self.poetry_run(
-            "install",
-            get_verb("--sync", self.activated_environment()),
+            *get_install_sync_command(self.activated_environment()),
             "--dry-run",
             capture_output=True,
         )
@@ -162,7 +193,9 @@ class PythonProject:
         )
 
     def virtual_environments(
-        self, *, create_default_if_missing: Union[bool, EnvironmentCreationBehavior] = False
+        self,
+        *,
+        create_default_if_missing: Union[bool, EnvironmentCreationBehavior] = False,
     ) -> Iterator[PythonEnvironment]:
         """The project's virtual environments. These are cached for performance.
 
@@ -183,16 +216,21 @@ class PythonProject:
         self, install: EnvironmentCreationBehavior = EnvironmentCreationBehavior.Full
     ) -> PythonEnvironment:
         """To be used only when no environments exist. Creates a default one by calling "poetry install"."""
+        command: list[str] = []
         if install is EnvironmentCreationBehavior.Full:
-            self.poetry_run("install")
+            command = self._generate_poetry_install_command()
         elif install is EnvironmentCreationBehavior.NoDev:
-            self.poetry_run("install", "--no-dev")
+            command = self._generate_poetry_install_command()
+            command.append("--no-dev")
+
+        if command:
+            self.poetry_run(*command)
         else:
             assert install is EnvironmentCreationBehavior.Empty
             try:
-                self.poetry_run("env", "use", "python3")
-            except CalledProcessError:
                 self.poetry_run("env", "use", "python")
+            except CalledProcessError:
+                self.poetry_run("env", "use", "python3")
 
         del self._virtual_environments_cache  # force cache refresh
         activated_environment = self.activated_environment()
@@ -211,8 +249,9 @@ class PythonProject:
             if (stripped := str_path.strip()) and (
                 match := re.fullmatch(ENVIRONMENT_PATH_PATTERN, stripped)
             ):
-                yield Path(match.groupdict()["path"].strip()), bool(
-                    match.groupdict().get("activated")
+                yield (
+                    Path(match.groupdict()["path"].strip()),
+                    bool(match.groupdict().get("activated")),
                 )
 
     def current_environment_belongs_to_project(self) -> bool:
@@ -227,20 +266,19 @@ class PythonProject:
 
     def bump(self) -> bool:
         """Bump (update) all dependencies to the lock file. Return True if changed."""
-        if not self.lock_path.exists():
+        if not self.poetry.locker.is_locked():
             return self.lock_if_needed()
 
-        content = self.lock_path.read_text()
+        content = self.poetry.locker.lock.read_text()
         self.poetry_run("update", "--lock", breakout_of_venv=True)
-        if content != self.lock_path.read_text():
+        self.poetry.locker._lock_data = None
+        if content != self.poetry.locker.lock.read_text():
             return True
         return False
 
     def build(self, target_path: Path = None) -> Path:
         """Builds the package's wheel. If a path is provided, it will be moved there.
         Returns final path to the artifact."""
-        # like coredump_detector-0.0.1-py3-none-any.whl
-        wheel_pattern = re.compile(r"(?P<distribution>\S+?)-(?P<version>.+?)-(?P<extra>.+)\.whl")
 
         if not self.activated_environment():
             # Starting from poetry 1.2.0, when calling `poetry build` without an environment setup, it will
@@ -248,16 +286,36 @@ class PythonProject:
             # Ensure that a default environment exists.
             self._create_default_poetry_install(install=EnvironmentCreationBehavior.Empty)
 
-        poetry_output = self.poetry_run("build", "--format", "wheel", capture_output=True)
-        wheel_match = wheel_pattern.search(poetry_output)
+        _ = self.poetry_run("build", "--format", "wheel", capture_output=True)
+
+        # cloudtrail_logs_firehose_ingest-0.1.0-py3-none-any.whl
+        expected_distribution = self.poetry.package.name.replace("-", "_").casefold()
+        expected_version = self.poetry.package.pretty_version
+        # like coredump_detector-0.0.1-py3-none-any.whl
+        filename_pattern = re.compile(
+            rf"(?P<distribution>{re.escape(expected_distribution)})-(?P<version>{expected_version})-.+\.whl"
+        )
+
+        # there's a file that matches the expected pattern in self.wheelhouse; find it.
+        wheel_match: Optional[re.Match[str]] = None
+        for file in (self.project_path / "dist").glob("*.whl"):
+            if file.is_file() and (match := filename_pattern.fullmatch(file.name)):
+                wheel_match = match
+                break
 
         if not wheel_match:
             raise StewException(
-                f"Unable able to find a wheel filename in poetry's output:\n{poetry_output}"
+                f"Could not find a wheel file for {self.poetry.package.pretty_name} "
+                f"version {self.poetry.package.pretty_version} in {self.project_path / 'dist'}."
             )
 
-        assert wheel_match["distribution"] == self.package.safe_name
-        assert wheel_match["version"] == str(self.package.version)
+        # validate
+        assert (
+            wheel_match.group("distribution").casefold() == expected_distribution
+        ), f"{wheel_match['distribution']} does not match {self.poetry.package.pretty_name}"
+        assert (
+            wheel_match["version"] == self.poetry.package.pretty_version
+        ), f"{wheel_match['version']} does not match {self.poetry.package.pretty_version}"
         wheel = (
             self.project_path / "dist" / Path(wheel_match.group())
         )  # group() gives the complete match
@@ -270,14 +328,14 @@ class PythonProject:
         target = target_path / wheel.name
         if not target_path.exists():
             target_path.mkdir(parents=True)
-        assert target_path.is_dir()
+        assert target_path.is_dir(), f"{target_path} could not be found from {Path('.').absolute()}"
         shutil.move(str(wheel), str(target))
 
         return target
 
     def export(self) -> str:
         """Generates the content of a `requirements.txt` file based on the lock."""
-        command = ["export"]
+        command = ["export", "--with-credentials"]
         if self.options.build_without_hashes:
             command.append("--without-hashes")
 
@@ -286,8 +344,8 @@ class PythonProject:
     def launch_continuous_integration(
         self,
         auto_fix: bool = False,
-        checks: Optional[List[str]] = None,
-        skips: Optional[List[str]] = None,
+        checks: Optional[Iterable[str]] = None,
+        skips: Optional[Iterable[str]] = None,
         quick: bool = False,
         parallel: bool = True,
         github: bool = False,
@@ -325,16 +383,29 @@ class PythonProject:
                 # return unless we are cleaning a non-cleaned environment
                 return
 
-        command = ["install"]
-        if sync:
-            command.append(get_verb("--sync", target_environment))
-        if quiet and not self.verbose:
-            command.append("--quiet")
-
+        command = self._generate_poetry_install_command(target_environment if sync else None, quiet)
         self.poetry_run(*command, environment=target_environment)
         target_environment.installed = True
         target_environment.cleaned |= sync
 
+    def _generate_poetry_install_command(
+        self, sync_target_environment: Optional[PythonEnvironment] = None, quiet: bool = False
+    ) -> List[str]:
+        command: List[str] = (
+            get_install_sync_command(sync_target_environment)
+            if sync_target_environment
+            else ["install"]
+        )
+
+        if quiet and not self.verbose:
+            command.append("--quiet")
+        if self.options.all_extras:
+            command.append("--all-extras")
+        elif self.options.extras:
+            for extra in self.options.extras:
+                command.extend(["--extras", extra])
+        return command
+
     def remove_egg_info(self) -> bool:
         """Removes the egg-info (editable project hook) from the folder. Returns True if we removed it."""
         if self.egg_path.exists():
@@ -356,7 +427,7 @@ class PythonProject:
 
     def lock_if_needed(self) -> bool:
         """Lock if needed, return True if ran."""
-        if not self.lock_path.exists() or self.lock_is_outdated():
+        if not (self.poetry.locker.is_locked() and self.poetry.locker.is_fresh()):
             self.poetry_run("lock", breakout_of_venv=True)
             return True
         return False
@@ -373,8 +444,6 @@ class PythonProject:
 
         The `environment` param will make that environment active (e.g.: `poetry env use` called before).
         """
-        global _CHECKED_POETRY_VERSION
-
         environment_variables = os.environ.copy()
         if breakout_of_venv:
             environment_variables.pop("VIRTUAL_ENV", None)
@@ -384,6 +453,7 @@ class PythonProject:
                 *find_python_tool(PythonTool.Poetry, environment=environment),
                 *commands,
                 "-vv" if self.verbose else "",
+                "--no-cache" if self.poetry.disable_cache else "",
                 working_directory=self.project_path,
                 capture_output=capture_output,
                 verbose=self.verbose,
@@ -413,4 +483,4 @@ class PythonProject:
                 self.poetry_run("env", "use", current_environment.python_executable)
 
     def __str__(self) -> str:
-        return f"{self.package.name} [{self.toml_path}]"
+        return f"{self.poetry.package.pretty_name} [{self.poetry.pyproject_path}]"
